#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#include <sqltypes.h>
#include <sql.h>
#include <sqlext.h>

#include <time.h>
#include <stdio.h>
#include <string.h>


#define ODBC_DSN "kxsrvmulti2_mysql5"
#define ODBC_USER "kxenodbc"
#define ODBC_PASSWORD "kxenodbc"

#define MAX_NB_FIELDS 1000

/* useful for customers
  We can stream the output to a file
  and with this bool still view the output */

static int gStdErr = 0;


#define CHECK_STMT_OR_RETURN(f, H, r) {\
		if(((r)!=SQL_SUCCESS )&& ((r) != SQL_SUCCESS_WITH_INFO)){displayStatementError(f, H, r);return H;}else{displayTrace( f,r );}}

#define CHECK_STMT(f,H, r) {\
		if((r)!=SQL_SUCCESS){displayStatementError(f, H, r);}else{displayTrace( f,r );Printf("\nno matter");}}

#define CHECK_CONN_OR_RETURN(f, H, r) {\
		if(((r)!=SQL_SUCCESS) && ((r) != SQL_SUCCESS_WITH_INFO)){displayConnectionError(f, H, r);return H;}else{displayTrace( f,r );}}

#define CHECK_CONN(f,H, r) {\
		if((hr!=SQL_SUCCESS){displayConnectionError(f, H, r;}else{displayTrace( f,r );Printf("\nno matter");}}

#define CHECK_ENV_OR_RETURN(f, H, r) {\
		if(((r)!=SQL_SUCCESS) && ((r) != SQL_SUCCESS_WITH_INFO)){displayEnvError(f, H, r);return H;}else{displayTrace( f,r );}}

#define CHECK_ENV(f,H, r) {\
		if((r)!=SQL_SUCCESS){displayEnvError(f, H, r;}else{displayTrace( f,r );Printf("\nno matter");}}



/*****************************************************************************/

void Printf(char* iFormat, 	... )
{
	static char lBuffer[ 2048 ]; /* Hum no check */

	va_list	marker;

	va_start( marker,iFormat );
	vsprintf( lBuffer,iFormat,marker );
	va_end(marker);
	fprintf( stdout,"%s",lBuffer );
	fflush( stdout );
	if( gStdErr )
	{
		fprintf( stderr,"%s",lBuffer );
		fflush( stderr );
	}
}

/*****************************************************************************/

void ShowTime( const char* iMess )
{
	char tmpbuf[1000] = {0};

	_strtime_s( tmpbuf, 128 );
	Printf( "\n%s:%s",tmpbuf,iMess );
}

/*****************************************************************************/

static int getline (SQLCHAR *s, const int lim)
{
fflush(stdin);
int c, i;

for (i=0; i < lim-1 && (c=getchar()) != EOF && c!='\n'; ++i)
        s[i] = (SQLCHAR)c;
s[i] = '\0';
fflush(stdin);
return i;
}

/*****************************************************************************/

static void displayTrace(const char* iFunction, SQLRETURN iReturnCode)
{
	Printf("\nFunction '%s' return '%d'",iFunction, (int)iReturnCode);
	fflush(stdout);
}

/*****************************************************************************/

static void displayStatementError(
	const char* iFunction,
	SQLHANDLE iStatement,
	SQLRETURN iReturnCode)
{
	Printf("\nFunction '%s' return '%d'",iFunction, (int)iReturnCode);
	SQLSMALLINT lMessageLength = 0;
	SQLSMALLINT lRecNumber = 1;
	SQLINTEGER lErrCode;
	RETCODE lRes = SQL_ERROR;
	SQLCHAR lErrStateBuff[ SQL_SQLSTATE_SIZE + 1 ] = { 0 };
	SQLCHAR lErrTextBuff[ SQL_MAX_MESSAGE_LENGTH + 1 ] = { 0 };

	while( lRes != SQL_SUCCESS && lRes != SQL_NO_DATA 
			&& lRes != SQL_INVALID_HANDLE )
	{
		*lErrStateBuff = 0;
		*lErrTextBuff = 0;
		lRes = SQLGetDiagRec(	
					SQL_HANDLE_STMT,
					iStatement,
					lRecNumber++,
					lErrStateBuff,
					&lErrCode,
					lErrTextBuff,
					SQL_MAX_MESSAGE_LENGTH-1,
					&lMessageLength );
		lErrStateBuff[ SQL_SQLSTATE_SIZE ] = 0;
		lErrTextBuff[ SQL_MAX_MESSAGE_LENGTH ] = 0;
		Printf( "\n[%s]:%s",lErrStateBuff,lErrTextBuff );
	}
	Printf("\n" );
	fflush( stdout );
}

/*****************************************************************************/

static void displayConnectionError(
	const char* iFunction,
	SQLHANDLE iConn,
	SQLRETURN iReturnCode)
{
	Printf("\nFunction '%s' return '%d'",iFunction, (int)iReturnCode);

	SQLSMALLINT lMessageLength = 0;
	SQLSMALLINT lRecNumber = 1;
	SQLINTEGER lErrCode;
	RETCODE lRes = SQL_ERROR;
	SQLCHAR lErrStateBuff[ SQL_SQLSTATE_SIZE + 1 ] = { 0 };
	SQLCHAR lErrTextBuff[ SQL_MAX_MESSAGE_LENGTH + 1 ] = { 0 };

	while( lRes != SQL_SUCCESS && lRes != SQL_NO_DATA 
			&& lRes != SQL_INVALID_HANDLE )
	{
		*lErrStateBuff = 0;
		*lErrTextBuff = 0;
		lRes = SQLGetDiagRec(	
					SQL_HANDLE_DBC,
					iConn,
					lRecNumber++,
					lErrStateBuff,
					&lErrCode,
					lErrTextBuff,
					SQL_MAX_MESSAGE_LENGTH-1,
					&lMessageLength );
		lErrStateBuff[ SQL_SQLSTATE_SIZE ] = 0;
		lErrTextBuff[ SQL_MAX_MESSAGE_LENGTH ] = 0;
		Printf( "\n[%s]:%s",lErrStateBuff,lErrTextBuff );
	}
	Printf("\n" );
	fflush( stdout );
}


/*****************************************************************************/

static void displayEnvError(
	const char* iFunction,
	SQLHANDLE iEnv,
	SQLRETURN iReturnCode)
{
	Printf("\nFunction '%s' return '%d'\n",iFunction, (int)iReturnCode);
	fflush( stdout );

	SQLSMALLINT lMessageLength = 0;
	SQLSMALLINT lRecNumber = 1;
	SQLINTEGER lErrCode;
	RETCODE lRes = SQL_ERROR;
	SQLCHAR lErrStateBuff[ SQL_SQLSTATE_SIZE + 1 ] = { 0 };
	SQLCHAR lErrTextBuff[ SQL_MAX_MESSAGE_LENGTH + 1 ] = { 0 };

	while( lRes != SQL_SUCCESS && lRes != SQL_NO_DATA 
			&& lRes != SQL_INVALID_HANDLE )
	{
		*lErrStateBuff = 0;
		*lErrTextBuff = 0;
		lRes = SQLGetDiagRec(	
					SQL_HANDLE_ENV,
					iEnv,
					lRecNumber++,
					lErrStateBuff,
					&lErrCode,
					lErrTextBuff,
					SQL_MAX_MESSAGE_LENGTH-1,
					&lMessageLength );
		lErrStateBuff[ SQL_SQLSTATE_SIZE ] = 0;
		lErrTextBuff[ SQL_MAX_MESSAGE_LENGTH ] = 0;
		Printf( "\n[%s]:%s",lErrStateBuff,lErrTextBuff );
	}
	Printf("\n" );
	fflush( stdout );
}

/*****************************************************************************/

static SQLHANDLE ODBC_Init( void )
{
    SQLRETURN 	lResult	= SQL_SUCCESS;
	SQLHANDLE lEnv = SQL_NULL_HENV;
	lResult = SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &lEnv);
	CHECK_ENV_OR_RETURN("SQLAllocHandle env", lEnv, lResult);
	lResult = SQLSetEnvAttr(lEnv,
							SQL_ATTR_ODBC_VERSION,
							(SQLPOINTER)SQL_OV_ODBC3,
							SQL_IS_UINTEGER);
	CHECK_ENV_OR_RETURN("SQLSetEnvAttr", lEnv, lResult);
	return lEnv;
}

/*****************************************************************************/

static SQLHANDLE ODBC_Connect( 
	SQLHANDLE iEnv,
	SQLCHAR *iODBC_DSN,
	SQLCHAR *iODBC_USER,
	SQLCHAR *iODBC_PASSWORD
 )
{
    SQLRETURN 	lResult	= SQL_SUCCESS;
	SQLHANDLE lConn = SQL_NULL_HDBC;

	lResult = SQLAllocHandle(SQL_HANDLE_DBC, iEnv, &lConn);
	CHECK_ENV_OR_RETURN("SQLAllocHandle connection", iEnv, lResult);

	lResult	= SQLConnect(lConn,
						 iODBC_DSN,
						 SQL_NTS,
						 iODBC_USER,
						 SQL_NTS,
						 iODBC_PASSWORD,
						 SQL_NTS);
	CHECK_CONN_OR_RETURN("SQLConnect", lConn, lResult);
	return lConn;

}

/*****************************************************************************/

static SQLHANDLE ODBC_CreateStatement(
	SQLHANDLE iConn )
{
	SQLRETURN lResult = SQL_SUCCESS;
	SQLHANDLE lStatement;

	lResult = SQLAllocHandle(SQL_HANDLE_STMT, iConn, &lStatement );
	CHECK_CONN_OR_RETURN("SQLAllocHandle statement", iConn,lResult);
	return  lStatement;
}

/*****************************************************************************/

static void ODBC_ResetStatement( SQLHANDLE iStatement )
{
	if( iStatement != SQL_NULL_HSTMT )
	{
		//SQLCancel( iStatement );
		SQLFreeStmt( iStatement,SQL_UNBIND );
		SQLFreeStmt( iStatement,SQL_RESET_PARAMS );
		SQLFreeStmt( iStatement,SQL_CLOSE );
	}
}

/*****************************************************************************/

static void ODBC_FreeStatement( SQLHANDLE& iStatement )
{
	if( iStatement != SQL_NULL_HSTMT )
	{
		SQLFreeStmt( iStatement,SQL_UNBIND );
		SQLFreeStmt( iStatement,SQL_RESET_PARAMS );
		SQLFreeStmt( iStatement,SQL_CLOSE );
		SQLFreeHandle( SQL_HANDLE_STMT,iStatement );
		iStatement = SQL_NULL_HSTMT;
	}
}

/*****************************************************************************/

static void ODBC_Disconnect( SQLHANDLE& oConn )
{
	if( oConn != SQL_NULL_HDBC )
	{
		SQLFreeHandle(SQL_HANDLE_DBC, oConn);
		oConn = SQL_NULL_HDBC;
	}
}




/*****************************************************************************/

SQLHANDLE ODBC_ListTables( SQLHANDLE iStatement )
{
    SQLRETURN 	lResult	= SQL_SUCCESS;
	Printf( "\nList of tables" );
	lResult = SQLTables(
		iStatement,
		NULL,
		0,
		(SQLCHAR*)"%",
		SQL_NTS,
		(SQLCHAR*)"%",
		SQL_NTS,
		(SQLCHAR*)"TABLE,VIEW",
		SQL_NTS );
	CHECK_STMT_OR_RETURN("SQLTables",iStatement, lResult);

	// transient
	SQLCHAR lTableName[ 256 ] = { 0 };
	SQLLEN lTableNameLength = 0;
	SQLCHAR lSchemaName[ 255 ] = { 0 };
	SQLLEN lSchemaNameLength;

	lResult = SQLBindCol(
					iStatement,
					(SQLUSMALLINT)2, // TABLE_SCHEM
					(SQLSMALLINT)SQL_C_CHAR,
					lSchemaName,
					(SQLLEN)254,
					&lSchemaNameLength );
	CHECK_STMT_OR_RETURN("SQLBindCol # 2 TABLE_SCHEM", iStatement, lResult);

	lResult = SQLBindCol(
					iStatement,
					(SQLUSMALLINT)3, // TABLE_NAME 
					(SQLSMALLINT)SQL_C_CHAR,
					lTableName,
					(SQLLEN)254,
					&lTableNameLength );
	CHECK_STMT_OR_RETURN("SQLBindCol # 3 TABLE_NAME", iStatement, lResult);

	lResult = SQLFetch( iStatement );

	Printf( "\n\nList of tables" );
	while( SQL_SUCCEEDED( lResult ) )
	{
		Printf( "%\n%s.%s",lSchemaName,lTableName );
		lResult = SQLFetch( iStatement );
	}
	Printf( "\n********************" );
	return iStatement;
}


/*****************************************************************************/

SQLHANDLE ODBC_ExistTable( 
	SQLHANDLE iStatement,
	const char * iSchemaName, const char * iTableName )
{
    SQLRETURN 	lResult	= SQL_SUCCESS;

	lResult = SQLTables(
		iStatement,
		NULL,
		0,
		NULL,
		0,
		(SQLCHAR*)iTableName,
		SQL_NTS,
		(SQLCHAR*)"TABLE,VIEW",
		SQL_NTS );
	CHECK_STMT_OR_RETURN("SQLTables",iStatement, lResult);
	lResult = SQLFetch( iStatement );
	if( lResult == SQL_SUCCESS || lResult == SQL_SUCCESS_WITH_INFO)
	{
		Printf( "\n%s exists",iTableName );
	}
	else
	{
		Printf( "\n%s does not exist",iTableName );
	}
	return iStatement;
}

/*****************************************************************************/

SQLHANDLE ODBC_ListSchemas( SQLHANDLE iStatement )
{
    SQLRETURN 	lResult	= SQL_SUCCESS;
	Printf( "\nRequestiong List of schemas" );
	lResult = SQLTables(
		iStatement,
		NULL,
		0,
		(SQLCHAR*)SQL_ALL_SCHEMAS ,
		1,
		(SQLCHAR*)NULL,
		0,
		(SQLCHAR*)NULL,
		0 );
	CHECK_STMT_OR_RETURN("SQLTables",iStatement, lResult);

	SQLCHAR lSchemaName[ 255 ] = { 0 };
	SQLLEN lSchemaNameLength;

	lResult = SQLBindCol(
					iStatement,
					(SQLUSMALLINT)2, // TABLE_SCHEM
					(SQLSMALLINT)SQL_C_CHAR,
					lSchemaName,
					(SQLLEN)254,
					&lSchemaNameLength );
	CHECK_STMT_OR_RETURN("SQLBindCol # 2 TABLE_SCHEM", iStatement, lResult);

	lResult = SQLFetch( iStatement );
	CHECK_STMT_OR_RETURN("SQLFetch ", iStatement, lResult);

	Printf( "\n\nList of schemas" );
	while( SQL_SUCCEEDED( lResult ) )
	{
		Printf( "%\n%s",lSchemaName );
		lResult = SQLFetch( iStatement );
	}
	Printf( "\n********************" );
	return iStatement;
}

/*****************************************************************************/

SQLHANDLE ODBC_PrepareTest( SQLHANDLE iStatement,const char * iTableName )
{
	ODBC_ResetStatement( iStatement );
	char lSQL[1024] = {0};

	SQLRETURN 	lResult	= SQL_SUCCESS;
	Printf( "\nDo some cleaning" );
	sprintf( lSQL,"DROP TABLE %s",iTableName );
	lResult	= SQLExecDirect( iStatement,(SQLCHAR*)lSQL,SQL_NTS );
	CHECK_STMT( lSQL,iStatement,lResult );

	Printf( "\nCreate a table with strange name : CREATE TABLE %s( ID INTEGER,DECIMAL10_0 DECIMAL(10,0),f FLOAT,DECIMAL10_1 DECIMAL(10,1))",iTableName );
	sprintf( lSQL,"CREATE TABLE %s ( ID1 INTEGER,DECIMAL10_0 DECIMAL(10,0),f FLOAT,DECIMAL10_1 DECIMAL(10,1))",iTableName );
	lResult	= SQLExecDirect( iStatement,(SQLCHAR*)lSQL,SQL_NTS );
	CHECK_STMT_OR_RETURN( lSQL,iStatement,lResult );
	return iStatement;
}

/*****************************************************************************/

SQLHANDLE ODBC_ListColumns( SQLHANDLE iStatement,const char * iTableName )
{
    SQLRETURN 	lResult	= SQL_SUCCESS;
	Printf( "\nRequesting List of columns of %s",iTableName );
	lResult = SQLColumns(
		iStatement,
		NULL,
		0,
		NULL,
		0,
		(SQLCHAR*)iTableName,
		SQL_NTS,
		NULL,
		0 );
	CHECK_STMT_OR_RETURN("SQLColumns",iStatement, lResult);

	SQLCHAR lColumnName[ 255 ] = { 0 };
	SQLLEN lColumnNameLength;
	SQLCHAR lNativeTypeName[ 255 ] = { 0 };
	SQLLEN lNativeTypeNameLength;

	lResult = SQLBindCol(
					iStatement,
					(SQLUSMALLINT)4, // COLUMN_NAME
					(SQLSMALLINT)SQL_C_CHAR,
					lColumnName,
					(SQLLEN)254,
					&lColumnNameLength );
	CHECK_STMT_OR_RETURN("SQLBindCol # 4 COLUMN_NAME", iStatement, lResult);

	lResult = SQLBindCol(
					iStatement,
					(SQLUSMALLINT)6, // TYPE_NAME
					(SQLSMALLINT)SQL_C_CHAR,
					lNativeTypeName,
					(SQLLEN)254,
					&lNativeTypeNameLength );
	CHECK_STMT_OR_RETURN("SQLBindCol # 6 TYPE_NAME", iStatement, lResult);

	lResult = SQLFetch( iStatement );
	CHECK_STMT_OR_RETURN("SQLFetch ", iStatement, lResult);

	Printf( "\nList of columns of %s",iTableName );
	while( SQL_SUCCEEDED( lResult ) )
	{
		Printf( "%\n%s %s",lColumnName,lNativeTypeName );
		lResult = SQLFetch( iStatement );
	}
	Printf( "\n********************" );
	return iStatement;
}

SQLHANDLE ODBC_DumpTable( SQLHANDLE iStatement,const char * iTableName,bool iDumpValues )
{
    SQLRETURN 	lResult	= SQL_SUCCESS;
	lResult = SQLColumns(	
							iStatement,
							NULL, // No catalog management
							0,
							NULL, // No schema
							0,
							(SQLCHAR*)iTableName,
							(SQLSMALLINT)SQL_NTS,
							NULL,0 );
	CHECK_STMT_OR_RETURN("SQLColumns", iStatement, lResult);

	// transient
	SQLCHAR lNativeColumnName[ 256 ] = { 0 };
	SQLLEN lNativeColumnNameLength = 0;
	SQLCHAR lNativeType[ 255 ] = { 0 };
	SQLLEN lNativeTypeLength;
	SQLSMALLINT lNativeTypeCode = 0;
	SQLULEN lColSize = 0;
	SQLSMALLINT lDecimalDigits = 0;
	SQLLEN lDontCare;
	
	// Some arrays to store permanently the infos we got from ODBC
	SQLCHAR lNativeColumnNames[MAX_NB_FIELDS][255] = { 0 };
	SQLSMALLINT lNativeTypeCodes[ MAX_NB_FIELDS ] = { 0 };
	SQLULEN lColSizes[ MAX_NB_FIELDS ] = { 0 };
	SQLSMALLINT lDecimalDigitss[ MAX_NB_FIELDS ] = { 0 };

	lResult = SQLBindCol(
					iStatement,
					(SQLUSMALLINT)4, // COLUMN_NAME
					(SQLSMALLINT)SQL_C_CHAR,
					lNativeColumnName,
					(SQLLEN)254,
					&lNativeColumnNameLength );
	CHECK_STMT_OR_RETURN("SQLBindCol # 4 COLUMN_NAME", iStatement, lResult);

	lResult = SQLBindCol(
					iStatement,
					(SQLUSMALLINT)6, // NATIVE_TYPE
					(SQLSMALLINT)SQL_C_CHAR,
					lNativeType,
					(SQLLEN)254,
					&lNativeTypeLength );
	CHECK_STMT_OR_RETURN("SQLBindCol # 6 NATIVE_TYPE", iStatement, lResult);

	lResult = SQLBindCol( 
				iStatement,
				(SQLUSMALLINT)5,  
				(SQLSMALLINT)SQL_C_SHORT,
				&lNativeTypeCode,
				(SQLLEN)sizeof( lNativeTypeCode ),
				&lDontCare );
	CHECK_STMT_OR_RETURN("SQLBindCol # 5 NATIVE_TYPE_CODE", iStatement, lResult);

	lResult = SQLBindCol( 
				iStatement,
				(SQLUSMALLINT)7,
				(SQLSMALLINT)SQL_C_LONG,
				&lColSize,
				(SQLLEN)sizeof( lColSize ),
				&lDontCare );
	CHECK_STMT_OR_RETURN("SQLBindCol # 6 COLSIZE", iStatement, lResult);

	lResult = SQLBindCol( 
				iStatement,
				(SQLUSMALLINT)9,
				(SQLSMALLINT)SQL_C_SHORT,
				&lDecimalDigits,
				(SQLLEN)sizeof( lDecimalDigits ),
				&lDontCare );
	CHECK_STMT_OR_RETURN("SQLBindCol # 9 DECIMAL_DIGITS", iStatement, lResult);

	lResult = SQLFetch( iStatement );
	int lNumField = 0;

	Printf( "\n\nHere is the MySQL description of table %s",iTableName );
	while( SQL_SUCCEEDED( lResult ) )
	{
		Printf( "\nField %d",lNumField );
		Printf( "\nNative Column Name %s",lNativeColumnName );
		Printf( "\nNative Type: %s",lNativeType );
		Printf( "\nNative Type Code %d",lNativeTypeCode );
		Printf( "\nCol Size %d",lColSize );
		Printf( "\nDecimal digits %d\n",lDecimalDigits );

		strcpy((char*)lNativeColumnNames[lNumField],(const char*)lNativeColumnName);
		lNativeTypeCodes[ lNumField ] = lNativeTypeCode;
		lColSizes[ lNumField ] = lColSize;
		lDecimalDigitss[ lNumField ] = lDecimalDigits;

		lNativeColumnName[ 0 ] = 0;
		lNativeType[ 0 ] = 0;
		lNativeTypeCode = 0;
		lColSize = 0;
		lDecimalDigits = 0;
		lResult = SQLFetch( iStatement );
		lNumField++;
	}
	ODBC_ResetStatement( iStatement );
	
	char lSelect[ 255 ] = { 0 };

	sprintf( lSelect,"SELECT * FROM %s",iTableName );
	lResult = SQLExecDirect( iStatement,(SQLCHAR*)lSelect,SQL_NTS );
	CHECK_STMT_OR_RETURN(lSelect, iStatement, lResult);

	int i = 0;
	Printf( "\n" );
	for( i = 0;i <lNumField;i++ )
	{
		Printf( "%s\t",lNativeColumnNames[i] );
	}
		
	char lValues[MAX_NB_FIELDS][255] = { 0 };
	SQLLEN lActualLengths[MAX_NB_FIELDS] = {0};

	for( i = 0;i <lNumField;i++ )
	{
		lResult	= SQLBindCol(	iStatement,
								(SQLUSMALLINT)i+1, 
								(SQLSMALLINT)SQL_C_CHAR,
								&(lValues[i][0]),
								(SQLLEN)lColSizes[i],
								&(lActualLengths[i]) );
		CHECK_STMT_OR_RETURN("SQLBindCol For Read", iStatement, lResult);
	}
	// Now we will fetch everything

	lResult = SQLFetch( iStatement );
	i = 0;
	while( SQL_SUCCEEDED(lResult) )
	{
		if( i % 100 == 0 )
		{
			Printf( "\n%d:", i );
		}
		if( iDumpValues )
		{
			int j = 0;

			Printf( "\n" );
			for(j=0;j < lNumField;j++ )
			{
				if(lActualLengths[j] == SQL_NULL_DATA )
				{
					Printf( "****\t" );
				}
				else
				{
					lValues[j][ lActualLengths[j]-1 ] = 0;
					Printf( "%s\t",lValues[j] );
				}
			}

		}
		i++;
		lResult = SQLFetch( iStatement );
	}
	Printf( "\n\n We have retrieved all data (%d rows )in table %s",i,iTableName );
	CHECK_STMT_OR_RETURN("End of retrieve data", iStatement, lResult);
	return iStatement;
}


/*****************************************************************************/

SQLHANDLE ODBC_QuickPrepareDump(
	SQLHANDLE iStatement,
	const char * iTableName,
	int& iNumField,
	char iNativeColumnNames[MAX_NB_FIELDS][255],
	char iValues[MAX_NB_FIELDS][255],
	SQLLEN iActualLengths[MAX_NB_FIELDS] )
{
    SQLRETURN 	lResult	= SQL_SUCCESS;


	
	char lSelect[ 255 ] = { 0 };

	sprintf( lSelect,"SELECT * FROM %s",iTableName );
	lResult = SQLExecDirect( iStatement,(SQLCHAR*)lSelect,SQL_NTS );
	CHECK_STMT_OR_RETURN(lSelect, iStatement, lResult);

	int i = 0;
	Printf( "\n" );
	for( i = 0;i <iNumField;i++ )
	{
		Printf( "%s\t",iNativeColumnNames[i] );
	}
		
	for( i = 0;i <iNumField;i++ )
	{
		lResult	= SQLBindCol(	iStatement,
								(SQLUSMALLINT)i+1, 
								(SQLSMALLINT)SQL_C_CHAR,
								&(iValues[i][0]),
								(SQLLEN)254,
								&(iActualLengths[i]) );
		CHECK_STMT_OR_RETURN("SQLBindCol For Read", iStatement, lResult);
	}
	return iStatement;
}

/*****************************************************************************/

SQLHANDLE ODBC_ExecuteSimpleSQL(
	SQLHANDLE iStatement,
	const char * iSQL )
{
    SQLRETURN 	lResult	= SQL_SUCCESS;

	lResult = SQLExecDirect( iStatement,(SQLCHAR*)iSQL,SQL_NTS );
	CHECK_STMT_OR_RETURN("SQLExecDirect ODBC_ExecuteSimpleSQL", iStatement, lResult);
}


/*****************************************************************************/

SQLHANDLE ODBC_PrepareDump(
	SQLHANDLE iStatement,
	const char * iTableName,
	int& iNumField,
	char iNativeColumnNames[MAX_NB_FIELDS][255],
	char iValues[MAX_NB_FIELDS][255],
	SQLLEN iActualLengths[MAX_NB_FIELDS] )
{
	ODBC_ResetStatement( iStatement );
    SQLRETURN 	lResult	= SQL_SUCCESS;
	lResult = SQLColumns(	
							iStatement,
							NULL, // No catalog management
							0,
							NULL, // No schema
							0,
							(SQLCHAR*)iTableName,
							(SQLSMALLINT)SQL_NTS,
							NULL,0 );
	CHECK_STMT_OR_RETURN("SQLColumns", iStatement, lResult);

	// transient
	SQLCHAR lNativeColumnName[ 256 ] = { 0 };
	SQLLEN lNativeColumnNameLength = 0;
	SQLCHAR lNativeType[ 255 ] = { 0 };
	SQLLEN lNativeTypeLength;
	SQLSMALLINT lNativeTypeCode = 0;
	SQLULEN lColSize = 0;
	SQLSMALLINT lDecimalDigits = 0;
	SQLLEN lDontCare;
	
	// Some arrays to store the infos we got from ODBC
	SQLSMALLINT lNativeTypeCodes[ MAX_NB_FIELDS ] = { 0 };
	SQLULEN lColSizes[ MAX_NB_FIELDS ] = { 0 };
	SQLSMALLINT lDecimalDigitss[ MAX_NB_FIELDS ] = { 0 };

	lResult = SQLBindCol(
					iStatement,
					(SQLUSMALLINT)4, // COLUMN_NAME
					(SQLSMALLINT)SQL_C_CHAR,
					lNativeColumnName,
					(SQLLEN)254,
					&lNativeColumnNameLength );
	CHECK_STMT_OR_RETURN("SQLBindCol # 4 COLUMN_NAME", iStatement, lResult);

	lResult = SQLBindCol(
					iStatement,
					(SQLUSMALLINT)6, // NATIVE_TYPE
					(SQLSMALLINT)SQL_C_CHAR,
					lNativeType,
					(SQLLEN)254,
					&lNativeTypeLength );
	CHECK_STMT_OR_RETURN("SQLBindCol # 6 NATIVE_TYPE", iStatement, lResult);

	lResult = SQLBindCol( 
				iStatement,
				(SQLUSMALLINT)5,  
				(SQLSMALLINT)SQL_C_SHORT,
				&lNativeTypeCode,
				(SQLLEN)sizeof( lNativeTypeCode ),
				&lDontCare );
	CHECK_STMT_OR_RETURN("SQLBindCol # 5 NATIVE_TYPE_CODE", iStatement, lResult);

	lResult = SQLBindCol( 
				iStatement,
				(SQLUSMALLINT)7,
				(SQLSMALLINT)SQL_C_LONG,
				&lColSize,
				(SQLLEN)sizeof( lColSize ),
				&lDontCare );
	CHECK_STMT_OR_RETURN("SQLBindCol # 6 COLSIZE", iStatement, lResult);

	lResult = SQLBindCol( 
				iStatement,
				(SQLUSMALLINT)9,
				(SQLSMALLINT)SQL_C_SHORT,
				&lDecimalDigits,
				(SQLLEN)sizeof( lDecimalDigits ),
				&lDontCare );
	CHECK_STMT_OR_RETURN("SQLBindCol # 9 DECIMAL_DIGITS", iStatement, lResult);

	lResult = SQLFetch( iStatement );
	iNumField = 0;

	Printf( "\n\nHere is the MySQL description of table %s",iTableName );
	while( SQL_SUCCEEDED( lResult ) )
	{
		Printf( "\nField %d",iNumField );
		Printf( "\nNative Column Name %s",lNativeColumnName );
		Printf( "\nNative Type: %s",lNativeType );
		Printf( "\nNative Type Code %d",lNativeTypeCode );
		Printf( "\nCol Size %d",lColSize );
		Printf( "\nDecimal digits %d\n",lDecimalDigits );

		strcpy((char*)iNativeColumnNames[iNumField],(const char*)lNativeColumnName);
		lNativeTypeCodes[ iNumField ] = lNativeTypeCode;
		lColSizes[ iNumField ] = lColSize;
		lDecimalDigitss[ iNumField ] = lDecimalDigits;

		lNativeColumnName[ 0 ] = 0;
		lNativeType[ 0 ] = 0;
		lNativeTypeCode = 0;
		lColSize = 0;
		lDecimalDigits = 0;
		lResult = SQLFetch( iStatement );
		iNumField++;
	}
	ODBC_ResetStatement( iStatement );
	
	char lSelect[ 255 ] = { 0 };

	sprintf( lSelect,"SELECT * FROM %s",iTableName );
	lResult = SQLExecDirect( iStatement,(SQLCHAR*)lSelect,SQL_NTS );
	CHECK_STMT_OR_RETURN(lSelect, iStatement, lResult);

	int i = 0;
	Printf( "\n" );
	for( i = 0;i <iNumField;i++ )
	{
		Printf( "%s\t",iNativeColumnNames[i] );
	}
		
	for( i = 0;i <iNumField;i++ )
	{
		lResult	= SQLBindCol(	iStatement,
								(SQLUSMALLINT)i+1, 
								(SQLSMALLINT)SQL_C_CHAR,
								&(iValues[i][0]),
								(SQLLEN)254,
								&(iActualLengths[i]) );
		CHECK_STMT_OR_RETURN("SQLBindCol For Read", iStatement, lResult);
	}
	return iStatement;
}

/*****************************************************************************/

SQLHANDLE ODBC_FetchFromPrepared( 
	bool iDumpValues,
	SQLHANDLE iStatement,
	const char * iTableName,
	int iNumField,
	char iNativeColumnNames[MAX_NB_FIELDS][255],
	char iValues[MAX_NB_FIELDS][255],
	SQLLEN iActualLengths[MAX_NB_FIELDS] )
{
    SQLRETURN 	lResult	= SQL_SUCCESS;

	char lSelect[ 255 ] = { 0 };
	int i = 0;
	Printf( "\n" );
	for( i = 0;i <iNumField;i++ )
	{
		Printf( "%s\t",iNativeColumnNames[i] );
	}
		
	// Now we will fetch everything

	lResult = SQLFetch( iStatement );
	i = 0;
	while( SQL_SUCCEEDED(lResult) )
	{
		if( i % 100 == 0 )
		{
			Printf( "\n%d:", i );
		}
	
		if( i == 8000 )
		{
			ShowTime( "We are sleeping a while at row 8000 just to simulate a very complex computation between two fetches\n");
			::Sleep(30*1000);
			ShowTime( "OK we restart to fetch\n");
		}
		if( iDumpValues )
		{
			int j = 0;

			Printf( "\n" );
			for(j=0;j < 1;j++ )
			{
				if(iActualLengths[j] == SQL_NULL_DATA )
				{
					Printf( "****\t" );
				}
				else
				{
					iValues[j][ iActualLengths[j]-1 ] = 0;
					Printf( "%s\t",iValues[j] );
				}
			}

		}
		i++;
		lResult = SQLFetch( iStatement );
	}
	Printf( "\n\n We have retrieved all data (%d rows )in table %s",i,iTableName );
	displayStatementError("After last fetch",iStatement,lResult);
	CHECK_STMT_OR_RETURN("End of retrieve data", iStatement, lResult);
	return iStatement;
}

SQLHANDLE ODBC_EndTran( SQLHANDLE iConnection )
{
	SQLRETURN lReturn = SQLEndTran( SQL_HANDLE_DBC,iConnection,SQL_COMMIT );
	CHECK_CONN_OR_RETURN("SQLEndTran", iConnection, lReturn);
}


			  
static char gNativeColumnNames1[MAX_NB_FIELDS][255] = {0};
static char gValues1[MAX_NB_FIELDS][255] = {0 };
static SQLLEN gActualLengths1[MAX_NB_FIELDS] = { 0 };

static char gNativeColumnNames2[MAX_NB_FIELDS][255] = {0};
static char gValues2[MAX_NB_FIELDS][255] = {0 };
static SQLLEN gActualLengths2[MAX_NB_FIELDS] = { 0 };

/*****************************************************************************/

int
main
(int argc, char* argv[])
{
	SQLCHAR lODBC_DSN[ 255 ] = ODBC_DSN;
	SQLCHAR lODBC_USER[ 255 ] = ODBC_USER;
	SQLCHAR lODBC_PASSWORD[ 255 ] = ODBC_PASSWORD;
	int i =0;

	for( i = argc - 1 ;i >0;i-- )
	{
		if( strcmp(argv[i],"-ask") == 0 )
		{
			Printf( "\nODBC DSN:");getline( lODBC_DSN,100 );
			Printf( "\nUser:" );getline( lODBC_USER,100 );
			Printf( "\nPassword:" );getline( lODBC_PASSWORD,100 );
			// Printf("\nSQL Request:");getline( lSQL_REQUEST,1900 );
			Printf( "\n");
		}
		if( strcmp(argv[i],"-stderr") == 0 )
		{
			gStdErr = 1 ;
		}
	}
	Printf( "\nConnexion:%s",lODBC_DSN);
	Printf( "\nUser:%s",lODBC_USER );
	Printf( "\nPassword:%s",lODBC_PASSWORD );

	SQLHANDLE lEnv = SQL_NULL_HENV;
	SQLHANDLE lConnection2 = SQL_NULL_HDBC;
	SQLHANDLE lConnection = SQL_NULL_HDBC;
	SQLHANDLE lStatement = SQL_NULL_HSTMT;
	SQLHANDLE lStatement2 = SQL_NULL_HSTMT;
	int lNumField = 0;

	lEnv = ODBC_Init();
	lConnection = ODBC_Connect( lEnv,lODBC_DSN,lODBC_USER,lODBC_PASSWORD );

	lStatement = ODBC_CreateStatement( lConnection );
	ODBC_ExecuteSimpleSQL( lStatement,"SET net_write_timeout=10" );
	Printf ("We prepare our test table\n");
	ODBC_ExecuteSimpleSQL( lStatement,"DROP TABLE ALAIN" );
	ODBC_ExecuteSimpleSQL( lStatement,"CREATE TABLE ALAIN( id INTEGER,fvalue float)" );
	ODBC_ExecuteSimpleSQL( lStatement,"INSERT INTO ALAIN VALUES( 1,1.1 )" );
	ODBC_ExecuteSimpleSQL( lStatement,"INSERT INTO ALAIN SELECT id+1,fvalue+1 FROM ALAIN" );
	ODBC_ExecuteSimpleSQL( lStatement,"INSERT INTO ALAIN SELECT id+2,fvalue+2 FROM ALAIN" );
	ODBC_ExecuteSimpleSQL( lStatement,"INSERT INTO ALAIN SELECT id+4,fvalue+4 FROM ALAIN" );
	ODBC_ExecuteSimpleSQL( lStatement,"INSERT INTO ALAIN SELECT id+8,fvalue+8 FROM ALAIN" );
	ODBC_ExecuteSimpleSQL( lStatement,"INSERT INTO ALAIN SELECT id+16,fvalue+16 FROM ALAIN" );
	ODBC_ExecuteSimpleSQL( lStatement,"INSERT INTO ALAIN SELECT id+32,fvalue+32 FROM ALAIN" );
	ODBC_ExecuteSimpleSQL( lStatement,"INSERT INTO ALAIN SELECT id+64,fvalue+64 FROM ALAIN" );
	ODBC_ExecuteSimpleSQL( lStatement,"INSERT INTO ALAIN SELECT id+128,fvalue+128 FROM ALAIN" );
	ODBC_ExecuteSimpleSQL( lStatement,"INSERT INTO ALAIN SELECT id+256,fvalue+256 FROM ALAIN" );
	ODBC_ExecuteSimpleSQL( lStatement,"INSERT INTO ALAIN SELECT id+512,fvalue+512 FROM ALAIN" );
	ODBC_ExecuteSimpleSQL( lStatement,"INSERT INTO ALAIN SELECT id+1024,fvalue+1024 FROM ALAIN" );
	ODBC_ExecuteSimpleSQL( lStatement,"INSERT INTO ALAIN SELECT id+2048,fvalue+2048 FROM ALAIN" );
	ODBC_ExecuteSimpleSQL( lStatement,"INSERT INTO ALAIN SELECT id+4096,fvalue+4096 FROM ALAIN" );
	ODBC_ExecuteSimpleSQL( lStatement,"INSERT INTO ALAIN SELECT id+8192,fvalue+8192 FROM ALAIN" );
	Printf ("Table ALAIN contains now 16384 rows\n");
	ODBC_PrepareDump( lStatement,"ALAIN",lNumField,gNativeColumnNames1,gValues1,gActualLengths1 );
	ODBC_FetchFromPrepared( false,lStatement,"ALAIN",lNumField,gNativeColumnNames1,gValues1,gActualLengths1 );
	Printf( "If you see that less than 16384 rows has been retrieved, you have reproduced the issue !!" );
	Printf( "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" );

	ODBC_Disconnect( lConnection ); 


	Printf( "\n\n" );
	return 0;
}
 
